"""Find, download and unpack wheels."""
import fnmatch
import hashlib
import itertools
import logging
import glob
import os
import re
import shutil
import yarg
import zipfile

from pathlib import Path
from requests_download import download, HashTracker
from tempfile import mkdtemp

from .util import get_cache_dir, normalize_path

logger = logging.getLogger(__name__)

class NoWheelError(Exception): pass

class CompatibilityScorer:
    """Score wheels for a given target platform

    0 for any score means incompatible.
    Higher numbers are more platform specific.
    """
    def __init__(self, py_version, platform):
        self.py_version = py_version
        py_version_tuple = tuple(map(int, py_version.split('.')[:2]))
        self.platform = platform
        # {('cp38', 'none', 'any'): N}  (higher N for more specific tags)
        self.tag_prio = {
            tag: i for i, tag in enumerate(reversed(
                list(compatible_tags(py_version_tuple, platform))
            ), start=1)
        }

    def score(self, whl_filename: str) -> int:
        """Return a number for how suitable a wheel is for the target Python

        Higher numbers mean more specific (preferred) tags. 0 -> incompatible.
        """
        m = re.search(r'-([^-]+)-([^-]+)-([^-]+)\.whl$', whl_filename)
        if not m:
            raise ValueError("Failed to find wheel tag in %r" % whl_filename)

        interpreter, abi, platform = m.group(1, 2, 3)
        # Expand compressed tags ('cp38.cp39' indicates compatibility w/ both)
        expanded_tags = itertools.product(
            interpreter.split('.'), abi.split('.'), platform.split('.')
        )
        return max(self.tag_prio.get(whl_tag, 0) for whl_tag in expanded_tags)

    def is_compatible(self, whl_filename: str) -> bool:
        return self.score(whl_filename) > 0

class WheelLocator(object):
    def __init__(self, requirement, scorer, extra_sources=None):
        self.requirement = requirement
        self.scorer = scorer
        self.extra_sources = extra_sources or []

        if requirement.count('==') != 1:
            raise ValueError("Requirement {!r} did not match name==version".format(requirement))
        self.name, self.version = requirement.split('==', 1)

    def pick_best_wheel(self, release_list):
        """Return the most specific compatible wheel

        Returns None if none of the supplied
        """
        best_score = 0
        best = None
        for release in release_list:
            if release.package_type != 'wheel':
                continue

            score = self.scorer.score(release.filename)
            if score == 0:
                # Incompatible
                continue

            if score > best_score:
                best = release
                best_score = score

        return best

    def check_extra_sources(self):
        """Find a compatible wheel in the specified extra_sources directories.

        Returns a Path or None.
        """
        whl_filename_prefix = '{name}-{version}-'.format(
            name=re.sub(r'[^\w\d.]+', '_', self.name),
            version=re.sub(r'[^\w\d.]+', '_', self.version),
        )
        for source in self.extra_sources:
            candidates = [CachedRelease(p.name)
                          for p in source.iterdir()
                          if p.name.startswith(whl_filename_prefix)]
            rel = self.pick_best_wheel(candidates)
            if rel:
                path = source / rel.filename
                return path

    def check_cache(self):
        """Find a wheel previously downloaded from PyPI in the cache.

        Returns a Path or None.
        """
        release_dir = get_cache_dir() / 'pypi' / self.name / self.version
        if not release_dir.is_dir():
            return None

        rel = self.pick_best_wheel(CachedRelease(p.name)
                                   for p in release_dir.iterdir())
        if rel is None:
            return None

        return release_dir / rel.filename

    def get_from_pypi(self):
        """Download a compatible wheel from PyPI.

        Downloads to the cache directory and returns the destination as a Path.
        Raises NoWheelError if no compatible wheel is found.
        """
        try:
            pypi_pkg = yarg.get(self.name)
        except yarg.HTTPError as e:
            if e.status_code == 404:
                raise NoWheelError("No package named {} found on PyPI".format(self.name))
            raise

        release_list = pypi_pkg.release(self.version)
        if release_list is None:
            raise NoWheelError("No release {0.version} for package {0.name}".format(self))

        preferred_release = self.pick_best_wheel(release_list)
        if preferred_release is None:
            raise NoWheelError('No compatible wheels found for {0.name} {0.version}'.format(self))

        download_to = get_cache_dir() / 'pypi' / self.name / self.version
        try:
            download_to.mkdir(parents=True)
        except OSError:
            # Ignore OSError only if the directory exists
            if not download_to.is_dir():
                raise
        target = download_to / preferred_release.filename

        from . import __version__
        hasher = HashTracker(hashlib.md5())
        headers = {'user-agent': 'pynsist/'+__version__}
        logger.info('Downloading wheel: %s', preferred_release.url)
        download(preferred_release.url, str(target), headers=headers,
                 trackers=(hasher,))
        if hasher.hashobj.hexdigest() != preferred_release.md5_digest:
            target.unlink()
            raise ValueError('Downloaded wheel corrupted: {}'.format(preferred_release.url))

        return target

    def fetch(self):
        """Find and return a compatible wheel (main interface)"""
        p = self.check_extra_sources()
        if p is not None:
            logger.info('Using wheel from extra directory: %s', p)
            return p

        p = self.check_cache()
        if p is not None:
            logger.info('Using cached wheel: %s', p)
            return p

        return self.get_from_pypi()


class CachedRelease(object):
    # Mock enough of the yarg Release object to be compatible with
    # pick_best_release above
    def __init__(self, filename):
        self.filename = filename
        self.package_type = 'wheel' if filename.endswith('.whl') else ''

def merge_dir_to(src, dst):
    """Merge all files from one directory into another.

    Subdirectories will be merged recursively. If filenames are the same, those
    from src will overwrite those in dst. If a regular file clashes with a
    directory, an error will occur.
    """
    for p in src.iterdir():
        if p.is_dir():
            dst_p = dst / p.name
            if dst_p.is_dir():
                merge_dir_to(p, dst_p)
            elif dst_p.is_file():
                raise RuntimeError('Directory {} clashes with file {}'
                                   .format(p, dst_p))
            else:
                shutil.copytree(str(p), str(dst_p))
        else:
            # Copy regular file
            dst_p = dst / p.name
            if dst_p.is_dir():
                raise RuntimeError('File {} clashes with directory {}'
                                   .format(p, dst_p))
            shutil.copy2(str(p), str(dst_p))


def extract_wheel(whl_file, target_dir, exclude=None):
    """Extract importable modules from a wheel to the target directory
    """
    # Extract to temporary directory
    td = Path(mkdtemp())
    with zipfile.ZipFile(str(whl_file), mode='r') as zf:
        if exclude:
            exclude_regexen = make_exclude_regexen(exclude)
            for zpath in zf.namelist():
                if is_excluded('pkgs/' + zpath, exclude_regexen):
                    continue  # Skip excluded paths
                zf.extract(zpath, path=str(td))
        else:
            zf.extractall(str(td))

    # Move extra lib files out of the .data subdirectory
    for p in td.iterdir():
        if p.suffix == '.data':
            if (p / 'purelib').is_dir():
                merge_dir_to(p / 'purelib', td)
            if (p / 'platlib').is_dir():
                merge_dir_to(p / 'platlib', td)

            # HACK: Some wheels from Christoph Gohlke's page have extra package
            # files added in data/Lib/site-packages. This is a trick that relies
            # on the default installation layout. It doesn't look like it will
            # change, so in the best tradition of packaging, we'll work around
            # the workaround.
            # https://github.com/takluyver/pynsist/issues/171
            # This is especially ugly because we do a case-insensitive match,
            # regardless of the filesystem.
            if (p / 'data').is_dir():
                for sd in (p / 'data').iterdir():
                    if sd.name.lower() == 'lib' and sd.is_dir():
                        for sd2 in sd.iterdir():
                            if sd2.name.lower() == 'site-packages' and sd2.is_dir():
                                merge_dir_to(sd2, td)

    # Copy to target directory
    target = Path(target_dir)
    copied_something = False
    for p in td.iterdir():
        if p.suffix not in {'.data'}:
            if p.is_dir():
                # If the dst directory already exists, this will combine them.
                # shutil.copytree will not combine them.
                try:
                    target.joinpath(p.name).mkdir()
                except OSError:
                    if not target.joinpath(p.name).is_dir():
                        raise
                merge_dir_to(p, target / p.name)
            else:
                shutil.copy2(str(p), str(target))
            copied_something = True

    if not copied_something:
        raise RuntimeError("Did not find any files to extract from wheel {}"
                           .format(whl_file))

    # Clean up temporary directory
    shutil.rmtree(str(td))

class WheelGetter:
    def __init__(self, requirements, wheel_globs, target_dir,
                 py_version, bitness, extra_sources=None, exclude=None):
        self.requirements = requirements
        self.wheel_globs = wheel_globs
        self.target_dir = target_dir
        target_platform = 'win_amd64' if bitness == 64 else 'win32'
        self.scorer = CompatibilityScorer(py_version, target_platform)
        self.extra_sources = extra_sources
        self.exclude = exclude

        self.got_distributions = {}

    def get_all(self):
        self.get_requirements()
        self.get_globs()

    def get_requirements(self):
        for req in self.requirements:
            wl = WheelLocator(req, self.scorer, self.extra_sources)
            whl_file = wl.fetch()
            extract_wheel(whl_file, self.target_dir, exclude=self.exclude)
            self.got_distributions[wl.name] = whl_file

    def get_globs(self):
        for glob_path in self.wheel_globs:
            paths = glob.glob(glob_path)
            if not paths:
                raise ValueError('Glob path {} does not match any files'
                                 .format(glob_path))
            for path in paths:
                logger.info('Collecting wheel file: %s (from: %s)',
                            os.path.basename(path), glob_path)
                self.validate_wheel(path)
                extract_wheel(path, self.target_dir, exclude=self.exclude)

    def validate_wheel(self, whl_path):
        """
        Verify that the given wheel can safely be included in the current installer.
        If so, the given wheel info will be included in the given wheel info array.
        If not, an exception will be raised.
        """
        wheel_name = os.path.basename(whl_path)
        distribution = wheel_name.split('-', 1)[0]

        # Check that a distribution of same name has not been included before
        if distribution in self.got_distributions:
            prev_path = self.got_distributions[distribution]
            raise ValueError('Multiple wheels specified for {}:\n  {}\n  {}'.format(
                             distribution, prev_path, whl_path))

        # Check that the wheel is compatible with the installer environment
        if not self.scorer.is_compatible(wheel_name):
            raise ValueError('Wheel {} is not compatible with Python {}, {}'
                .format(wheel_name, self.scorer.py_version, self.scorer.platform))

        self.got_distributions[distribution] = whl_path


def make_exclude_regexen(exclude_patterns):
    """Translate exclude glob patterns to regex pattern objects.

    Handles matching files under a named directory.
    """
    re_pats = set()
    for pattern in exclude_patterns:
        re_pats.add(fnmatch.translate(pattern))
        if not pattern.endswith('*'):
            # Also use the pattern as a directory name and match anything
            # under that directory.
            suffix = '*' if pattern.endswith('/') else '/*'
            re_pats.add(fnmatch.translate(pattern + suffix))

    return [re.compile(p) for p in sorted(re_pats)]


def is_excluded(path, exclude_regexen):
    """Return True if path matches an exclude pattern"""
    path = normalize_path(path)
    for re_pattern in exclude_regexen:
        if re_pattern.match(path):
            return True
    return False

# The function below is based on the packaging.tags module, used with
# modification following the BSD 2 clause license:

# Copyright (c) Donald Stufft and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

def compatible_tags(python_version : tuple =None, platform : str =None):
    """Iterate through compatible tags for our target Python

    Tags are yielded in order from the most specific to the most general.

    Based on packaging.tags module, but simplified for Pynsist's use case,
    and avoiding getting any details from the currently running Python.
    """
    interpreter = "cp{}{}".format(python_version[0], python_version[1])

    cpython_abi = interpreter
    # Python is normally built with the pymalloc (m) option, and most wheels
    # are published for this ABI. The flag is dropped in Python 3.8.
    if python_version < (3, 8):
        cpython_abi += 'm'

    yield interpreter, cpython_abi, platform
    yield interpreter, "abi3", platform
    yield interpreter, "none", platform

    # cp3x-abi3 down to cp32 (Python 3.2 was the first version to have ABI3)
    for minor_version in range(python_version[1] - 1, 1, -1):
        interpreter = "cp{}{}".format(python_version[0], minor_version)
        yield interpreter, "abi3", platform

    py_interps = [
        f"py{python_version[0]}{python_version[1]}",  # e.g. py38
        f"py{python_version[0]}",                     # py3
    ] + [
        f"py{python_version[0]}{minor}"               # py37 ... py30
        for minor in range(python_version[1] - 1, -1, -1)
    ]

    for version in py_interps:
        yield version, "none", platform
    for version in py_interps:
        yield version, "none", "any"



import os
import nsist

example_cfgs = [
    'console/installer.cfg',
    'pygame/installer.cfg',
    # 'pygi_mpl_numpy/installer.cfg',  # this example does not currently work
    'pyglet/installer.cfg',
    'pyqt5/installer.cfg',
    'pyqt5_opencv/installer.cfg',
    'pyqt5_qml/installer.cfg',
    'pywebview/installer.cfg',
]

examples_dir = os.path.dirname(os.path.abspath(__file__))

for example_cfg in example_cfgs:
    os.chdir(examples_dir)
    nsist.main([example_cfg])



import os
import sys
import time
import threading
from PyQt5 import QtCore, QtGui, QtQml


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class JSManager(QtCore.QObject):
    text = QtCore.pyqtSignal(QtCore.QVariant)

    @QtCore.pyqtSlot()
    def get_text(self):
        def go():
            _text = "Hello from JavaScript"
            self.text.emit(_text)
        threading.Thread(target=go).start()


def main():
    app = QtGui.QGuiApplication(sys.argv)
    QtQml.qmlRegisterType(JSManager, 'JSManager', 1, 0, 'JSManager')
    engine = QtQml.QQmlApplicationEngine(os.path.join(THIS_DIR, "main.qml"))
    app.exec_()



import os
from subprocess import Popen, PIPE, STDOUT
import sys
import time
import webbrowser


def main():

    # Getting path to python executable (full path of deployed python on Windows)
    executable = sys.executable

    path_to_main = os.path.join(os.path.dirname(__file__), "app.py")

    # Running streamlit server in a subprocess and writing to log file
    proc = Popen(
        [
            executable,
            "-m",
            "streamlit",
            "run",
            path_to_main,
            # The following option appears to be necessary to correctly start the streamlit server,
            # but it should start without it. More investigations should be carried out.
            "--server.headless=true",
            "--global.developmentMode=false",
        ],
        stdin=PIPE,
        stdout=PIPE,
        stderr=STDOUT,
        text=True,
    )
    proc.stdin.close()

    # Force the opening (does not open automatically) of the browser tab after a brief delay to let
    # the streamlit server start.
    time.sleep(3)
    webbrowser.open("http://localhost:8501")

    while True:
        s = proc.stdout.read()
        if not s:
            break
        print(s, end="")

    proc.wait()


if __name__ == "__main__":
    main()



import io
import os
from os.path import join as pjoin
from testpath import assert_isfile

from nsist import InstallerBuilder, DEFAULT_ICON
from .utils import test_dir


sample_preamble = pjoin(test_dir, u'sample_preamble.py')

def test_prepare_shortcuts(tmpdir):
    tmpdir = str(tmpdir)
    shortcuts = {'sc1': {'entry_point': 'norwegian.blue:parrot',
                         'icon': DEFAULT_ICON,
                         'console': False,
                         'extra_preamble': sample_preamble}}
    ib = InstallerBuilder("Test App", "1.0", shortcuts, build_dir=tmpdir)
    ib.prepare_shortcuts()

    scfile = pjoin(tmpdir, 'sc1.launch.pyw')
    assert_isfile(scfile)

    with io.open(scfile, 'r', encoding='utf-8') as f:
        contents = f.read()

    last2lines = [l.strip() for l in contents.rstrip().splitlines()[-2:]]
    assert last2lines == ['from norwegian.blue import parrot', 'parrot()']

    with io.open(sample_preamble, 'r', encoding='utf-8') as f:
        preamble_contents = f.read().strip()

    assert preamble_contents in contents

def test_copy_extra_files(tmpdir):
    tmpdir = str(tmpdir)
    files = [
        (pjoin(test_dir, 'data_files', 'dir1', 'eg-data.txt'), '$INSTDIR'),
        (pjoin(test_dir, 'data_files', 'dir2', 'eg-data.txt'), '$INSTDIR\\foo'),
        (pjoin(test_dir, 'data_files', 'dir1', 'subdir'), '$INSTDIR'),
        (pjoin(test_dir, 'data_files', 'dir2', 'subdir'), '$INSTDIR\\foo'),
    ]
    ib = InstallerBuilder("Test App", "1.0", {}, extra_files=files,
                          build_dir=tmpdir)
    ib.copy_extra_files()

    build_dir_files = set(os.listdir(tmpdir))
    for file in ['eg-data.txt', 'eg-data.1.txt', 'subdir', 'subdir.1']:
        assert file in build_dir_files

    assert ib.install_dirs == [
        ('subdir', '$INSTDIR'),
        ('subdir.1', '$INSTDIR\\foo'),
    ]
    assert ib.install_files == [
        ('eg-data.txt', '$INSTDIR'),
        ('eg-data.1.txt', '$INSTDIR\\foo'),
    ]

def test_copy_installer_nsi(tmpdir):
    tmpdir = str(tmpdir)
    files = [
        (pjoin(test_dir, 'data_files', 'dir1', 'installer.nsi'), None),
    ]
    ib = InstallerBuilder("Test App", "1.0", {}, extra_files=files,
                          build_dir=tmpdir)
    ib.copy_extra_files()

    assert_isfile(pjoin(tmpdir, 'installer.1.nsi'))
    assert ib.install_files == [('installer.1.nsi', '$INSTDIR')]



#!/usr/bin/env python
# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions 
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

'''Bounces balls around a window and plays noises.

This is a simple demonstration of how pyglet efficiently manages many sound
channels without intervention.
'''

import os
import random
import sys

from pyglet.gl import *
import pyglet
from pyglet.window import key

pyglet.resource.path.insert(0, os.path.dirname(__file__))
pyglet.resource.reindex()
BALL_IMAGE = 'ball.png'
BALL_SOUND = 'ping.wav'

sound = pyglet.resource.media(BALL_SOUND, streaming=False)

class Ball(pyglet.sprite.Sprite):
    ball_image = pyglet.resource.image(BALL_IMAGE)
    width = ball_image.width
    height = ball_image.height

    def __init__(self):
        x = random.random() * (window.width - self.width)
        y = random.random() * (window.height - self.height)

        super(Ball, self).__init__(self.ball_image, x, y, batch=balls_batch)

        self.dx = (random.random() - 0.5) * 1000
        self.dy = (random.random() - 0.5) * 1000

    def update(self, dt):
        if self.x <= 0 or self.x + self.width >= window.width:
            self.dx *= -1
            sound.play()
        if self.y <= 0 or self.y + self.height >= window.height:
            self.dy *= -1
            sound.play()
        self.x += self.dx * dt
        self.y += self.dy * dt

        self.x = min(max(self.x, 0), window.width - self.width)
        self.y = min(max(self.y, 0), window.height - self.height)

window = pyglet.window.Window(640, 480, caption='Noisy')

@window.event
def on_key_press(symbol, modifiers):
    if symbol == key.SPACE:
        balls.append(Ball())
    elif symbol == key.BACKSPACE:
        if balls:
            del balls[-1]
    elif symbol == key.ESCAPE:
        window.has_exit = True

@window.event
def on_draw():
    window.clear()
    balls_batch.draw()
    label.draw()

def update(dt):
    for ball in balls:
        ball.update(dt)

balls_batch = pyglet.graphics.Batch()
label = pyglet.text.Label('Press space to add a ball, backspace to remove',
                          font_size=14,
                          x=window.width // 2, y=10, 
                          anchor_x='center')

balls = []

def main():
    pyglet.clock.schedule_interval(update, 1/30.)
    pyglet.app.run()



# (c) Continuum Analytics, Inc. / http://continuum.io
# All Rights Reserved
# Copied from conda constructor at commit d91adfb1c49666768ef9fd625d02276af6ddb0c9
# This file is under the BSD license:
#
# Copyright (c) 2016, Continuum Analytics, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Continuum Analytics, Inc. nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CONTINUUM ANALYTICS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# Helper script for adding and removing entries in the
# Windows system path from the NSIS installer.

__all__ = ['remove_from_system_path', 'add_to_system_path', 'broadcast_environment_settings_change']

import sys
import os, ctypes
from os import path
from ctypes import wintypes
if sys.version_info[0] >= 3:
    import winreg as reg
else:
    import _winreg as reg

HWND_BROADCAST = 0xffff
WM_SETTINGCHANGE = 0x001A
SMTO_ABORTIFHUNG = 0x0002
SendMessageTimeout = ctypes.windll.user32.SendMessageTimeoutW
SendMessageTimeout.restype = None #wintypes.LRESULT
SendMessageTimeout.argtypes = [wintypes.HWND, wintypes.UINT, wintypes.WPARAM,
            wintypes.LPCWSTR, wintypes.UINT, wintypes.UINT, ctypes.POINTER(wintypes.DWORD)]

def sz_expand(value, value_type):
    if value_type == reg.REG_EXPAND_SZ:
        return reg.ExpandEnvironmentStrings(value)
    else:
        return value

def remove_from_system_path(pathname, allusers=True, path_env_var='PATH'):
    """Removes all entries from the path which match the value in 'pathname'

       You must call broadcast_environment_settings_change() after you are finished
       manipulating the environment with this and other functions.

       For example,
         # Remove Anaconda from PATH
         remove_from_system_path(r'C:\Anaconda')
         broadcast_environment_settings_change()
    """
    pathname = path.normcase(path.normpath(pathname))

    envkeys = [(reg.HKEY_CURRENT_USER, r'Environment')]
    if allusers:
        envkeys.append((reg.HKEY_LOCAL_MACHINE,
            r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment'))
    for root, keyname in envkeys:
        key = reg.OpenKey(root, keyname, 0,
                reg.KEY_QUERY_VALUE|reg.KEY_SET_VALUE)
        reg_value = None
        try:
            reg_value = reg.QueryValueEx(key, path_env_var)
        except WindowsError:
            # This will happen if we're a non-admin install and the user has
            # no PATH variable.
            reg.CloseKey(key)
            continue

        try:
            any_change = False
            results = []
            for v in reg_value[0].split(os.pathsep):
                vexp = sz_expand(v, reg_value[1])
                # Check if the expanded path matches the
                # requested path in a normalized way
                if path.normcase(path.normpath(vexp)) == pathname:
                    any_change = True
                else:
                    # Append the original unexpanded version to the results
                    results.append(v)

            modified_path = os.pathsep.join(results)
            if any_change:
                reg.SetValueEx(key, path_env_var, 0, reg_value[1], modified_path)
        except:
            # If there's an error (e.g. when there is no PATH for the current
            # user), continue on to try the next root/keyname pair
            reg.CloseKey(key)

def add_to_system_path(paths, allusers=True, path_env_var='PATH'):
    """Adds the requested paths to the system PATH variable.

       You must call broadcast_environment_settings_change() after you are finished
       manipulating the environment with this and other functions.

    """
    # Make sure it's a list
    if not issubclass(type(paths), list):
        paths = [paths]

    # Ensure all the paths are valid before we start messing with the
    # registry.
    new_paths = None
    for p in paths:
        p = path.abspath(p)
        if not path.isdir(p):
            raise RuntimeError(
                'Directory "%s" does not exist, '
                'cannot add it to the path' % p
            )
        if new_paths:
            new_paths = new_paths + os.pathsep + p
        else:
            new_paths = p

    if allusers:
        # All Users
        root, keyname = (reg.HKEY_LOCAL_MACHINE,
            r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment')
    else:
        # Just Me
        root, keyname = (reg.HKEY_CURRENT_USER, r'Environment')

    key = reg.OpenKey(root, keyname, 0,
            reg.KEY_QUERY_VALUE|reg.KEY_SET_VALUE)

    reg_type = None
    reg_value = None
    try:
        try:
            reg_value = reg.QueryValueEx(key, path_env_var)
        except WindowsError:
            # This will happen if we're a non-admin install and the user has
            # no PATH variable; in which case, we can write our new paths
            # directly.
            reg_type = reg.REG_EXPAND_SZ
            final_value = new_paths
        else:
            reg_type = reg_value[1]
            # If we're an admin install, put us at the end of PATH.  If we're
            # a user install, throw caution to the wind and put us at the
            # start.  (This ensures we're picked up as the default python out
            # of the box, regardless of whether or not the user has other
            # pythons lying around on their PATH, which would complicate
            # things.  It's also the same behavior used on *NIX.)
            if allusers:
                final_value = reg_value[0] + os.pathsep + new_paths
            else:
                final_value = new_paths + os.pathsep + reg_value[0]

        reg.SetValueEx(key, path_env_var, 0, reg_type, final_value)

    finally:
        reg.CloseKey(key)

def broadcast_environment_settings_change():
    """Broadcasts to the system indicating that master environment variables have changed.

    This must be called after using the other functions in this module to
    manipulate environment variables.
    """
    SendMessageTimeout(HWND_BROADCAST, WM_SETTINGCHANGE, 0, u'Environment',
                SMTO_ABORTIFHUNG, 5000, ctypes.pointer(wintypes.DWORD()))


def main():
    if len(sys.argv) < 3:
        sys.exit("Too few arguments: {}".format(sys.argv))
    elif len(sys.argv) > 3:
        sys.exit("Too many arguments: {}".format(sys.argv))

    if sys.argv[1] == 'add':
        add_to_system_path(sys.argv[2])
    elif sys.argv[1] == 'add_user':
        add_to_system_path(sys.argv[2], allusers=False)
    elif sys.argv[1] == 'remove':
        remove_from_system_path(sys.argv[2])
    elif sys.argv[1] == 'remove_user':
        remove_from_system_path(sys.argv[2], allusers=False)

    broadcast_environment_settings_change()

if __name__ == '__main__':
    main()



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QObject
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QImage
import cv2


class CameraDevice(QObject):

    frame_ready = pyqtSignal(QImage)

    def __init__(self, device_id=0):
        super().__init__()
        self.capture = cv2.VideoCapture(device_id)
        self.timer = QTimer()

        if not self.capture.isOpened():
            raise ValueError("Device not found")

        self.timer.timeout.connect(self.read_frame)
        self.timer.setInterval(1000 / (self.fps or 30))
        self.timer.start()

    def __del__(self):
        self.timer.stop()
        self.capture.release()

    @property
    def fps(self):
        """Frames per second."""
        return int(self.capture.get(cv2.CAP_PROP_FPS))

    @property
    def size(self):
        """Returns the size of the video frames: (width, height)."""
        width = int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        return (width, height)

    def read_frame(self):
        """Read frame into QImage and emit it."""
        success, frame = self.capture.read()
        if success:
            img = _convert_array_to_qimage(frame)
            self.frame_ready.emit(img)
        else:
            raise ValueError("Failed to read frame")


def _convert_array_to_qimage(a):
    height, width, channels = a.shape
    bytes_per_line = channels * width
    cv2.cvtColor(a, cv2.COLOR_BGR2RGB, a)
    return QImage(a.data, width, height, bytes_per_line, QImage.Format_RGB888)



"""
This example demonstrates how to create a webview window.

Adapted from a pywebview example:
https://github.com/r0x0r/pywebview/blob/master/examples/simple_browser.py
"""

import webview

def main():
    # Create a non-resizable webview window with 800x600 dimensions
    webview.create_window("Simple browser", "http://pynsist.readthedocs.io/",
                            width=800, height=600, resizable=False)

if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""
This test program utilizes the Python bindings for GTK3 (PyGI or PyGObject).

The program consists of a window with a button that closes the window when
clicked. The window also shows a matplotlib plot. The plot is based on this
example: http://matplotlib.org/examples/pie_and_polar_charts/polar_bar_demo.html
"""

from gi.repository import Gtk
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.cm as cm
import numpy as np


class MyWindow(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self, title="PyGI / PyGObject Example")
        self.connect("delete-event", Gtk.main_quit)
        self.set_default_size(400, 400)

        self.box = Gtk.Box(spacing=6, orientation=Gtk.Orientation.VERTICAL)
        self.add(self.box)

        self.fig = Figure(figsize=(10,10), dpi=80)
        self.ax = self.fig.add_subplot(111, polar=True)

        N = 20
        theta = np.linspace(0.0, 2 * np.pi, N, endpoint=False)
        radii = 10 * np.random.rand(N)
        width = np.pi / 4 * np.random.rand(N)
        bars = self.ax.bar(theta, radii, width=width, bottom=0.0)
        for r, bar in zip(radii, bars):
            bar.set_facecolor(cm.jet(r / 10.))
            bar.set_alpha(0.5)

        self.canvas = FigureCanvas(self.fig)
        self.box.pack_start(self.canvas, True, True, 0)

        self.button = Gtk.Button(label="Exit")
        self.button.connect("clicked", self.on_exit_clicked)
        self.box.pack_start(self.button, False, False, 0)

    def on_exit_clicked(self, widget):
        Gtk.main_quit()

def main():
    win = MyWindow()
    win.show_all()
    Gtk.main()



import streamlit as st
import plotly.graph_objects as go

st.title("My first Streamlit app deployed with pynsist!")

st.write("Example plotly graphs follows")

fig = go.Figure(go.Scatter(x=[1, 2], y=[1, 2], name="trace", mode="lines"))
st.plotly_chart(fig, use_container_width=True)




import glob
import os
from pathlib import Path
import platform
import subprocess
from zipfile import ZipFile

import pytest
from testpath import assert_isfile, assert_isdir, assert_not_path_exists

from nsist.wheels import WheelGetter, extract_wheel

# To exclude tests requiring network on an unplugged machine, use: pytest -m "not network"

@pytest.mark.network
def test_matching_one_pattern(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    subprocess.call(['pip', 'wheel', 'requests==2.19.1', '-w', str(td1)])

    wg = WheelGetter([], [os.path.join(td1, '*.whl')], td2, platform.python_version(), 64)
    wg.get_globs()

    assert_isdir(os.path.join(td2, 'requests'))
    assert_isfile(os.path.join(td2, 'requests-2.19.1.dist-info', 'METADATA'))

    assert_isdir(os.path.join(td2, 'urllib3'))
    assert glob.glob(os.path.join(td2, 'urllib3*.dist-info'))

@pytest.mark.network
def test_duplicate_wheel_files_raise(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    subprocess.call(['pip', 'wheel', 'requests==2.19.1', '-w', str(td1)])

    wg = WheelGetter(['requests==2.19.1'], [os.path.join(td1, '*.whl')], td2,
                     platform.python_version(), 64)

    with pytest.raises(ValueError, match='Multiple wheels specified'):
        wg.get_all()

def test_invalid_wheel_file_raise(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    open(os.path.join(td1, 'notawheel.txt'), 'w+')

    wg = WheelGetter([], [os.path.join(td1, '*')], td2,
                     platform.python_version(), 64)

    with pytest.raises(ValueError, match='notawheel.txt'):
        wg.get_globs()

def test_incompatible_platform_wheel_file_raise(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    Path(td1, 'incompatiblewheel-1.0.0-py2.py3-none-linux_x86_64.whl').touch()

    wg = WheelGetter([], [os.path.join(td1, '*.whl')], td2,
                     platform.python_version(), 64)

    with pytest.raises(ValueError, match='not compatible with .* win_amd64'):
        wg.get_globs()

def test_incompatible_python_wheel_file_raise(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    Path(td1, 'incompatiblewheel-1.0.0-py26-none-any.whl').touch()

    wg = WheelGetter([], [os.path.join(td1, '*.whl')], td2,
                     platform.python_version(), 64)

    with pytest.raises(ValueError, match='not compatible with Python {}'
                       .format(platform.python_version())):
        wg.get_globs()

def test_useless_wheel_glob_path_raise(tmpdir):
    td1 = str(tmpdir.mkdir('wheels'))
    td2 = str(tmpdir.mkdir('pkgs'))

    wg = WheelGetter([], [os.path.join(td1, '*.whl')], td2, '3.6', 64)

    with pytest.raises(ValueError, match='does not match any files'):
        wg.get_globs()

def test_extract_exclude_folder(tmpdir):
    whl_file = str(tmpdir / 'foo.whl')
    pkgs = tmpdir.mkdir('pkgs')

    with ZipFile(whl_file, 'w') as zf:
        zf.writestr('foo/bar.txt', b'blah')
        zf.writestr('foo/bar/abc.txt', b'blah')

    extract_wheel(whl_file, str(pkgs), exclude=['pkgs/foo/bar'])

    assert_isfile(str(pkgs / 'foo' / 'bar.txt'))
    assert_not_path_exists(str(pkgs / 'foo' / 'bar'))

def test_extract_data_lib_sitepkg(tmpdir):
    whl_file = str(tmpdir / 'foo.whl')
    pkgs = tmpdir.mkdir('pkgs')

    with ZipFile(whl_file, 'w') as zf:
        zf.writestr('osgeo/bar.txt', b'blah')
        # The case of 'Lib/site-packages' shouldn't matter
        zf.writestr('foo-1.0.data/data/Lib/siTE-packages/osgeo/abc.txt', b'a')
        zf.writestr('foo-1.0.data/data/lib/site-packages/osgeo/def.txt', b'd')

    extract_wheel(whl_file, str(pkgs), exclude=['pkgs/foo/bar'])

    assert_isfile(str(pkgs / 'osgeo' / 'bar.txt'))
    assert_isfile(str(pkgs / 'osgeo' / 'abc.txt'))
    assert_isfile(str(pkgs / 'osgeo' / 'def.txt'))



import os
import logging
from pathlib import Path
import requests
import sys

logger = logging.getLogger(__name__)


def download(url, target):
    """Download a file using requests.
    
    This is like urllib.request.urlretrieve, but requests validates SSL
    certificates by default.
    """
    if isinstance(target, Path):
        target = str(target)

    from . import __version__
    headers = {'user-agent': 'Pynsist/'+__version__}
    r = requests.get(url, headers=headers, stream=True)
    r.raise_for_status()
    with open(target, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024): 
            if chunk:
                f.write(chunk)

CACHE_ENV_VAR = 'PYNSIST_CACHE_DIR'

def get_cache_dir(ensure_existence=False):
    specified = os.environ.get(CACHE_ENV_VAR, None)

    if specified:
        p = Path(specified)
    elif os.name == 'posix' and sys.platform != 'darwin':
        # Linux, Unix, AIX, etc.
        # use ~/.cache if empty OR not set
        xdg = os.environ.get("XDG_CACHE_HOME", None) or (os.path.expanduser('~/.cache'))
        p = Path(xdg, 'pynsist')

    elif sys.platform == 'darwin':
        p = Path(os.path.expanduser('~'), 'Library/Caches/pynsist')

    else:
        # Windows (hopefully)
        local = os.environ.get('LOCALAPPDATA', None) or (os.path.expanduser('~\\AppData\\Local'))
        if local.startswith('~'):
            logger.warning("Could not find cache directory. Please set any of "
                           "these environment variables: "
                           "LOCALAPPDATA, HOME, USERPROFILE or HOMEPATH")
        p = Path(local, 'pynsist')

    if ensure_existence:
        p.mkdir(parents=True, exist_ok=True)

    return p


def normalize_path(path):
    """Normalize paths to contain "/" only"""
    return os.path.normpath(path).replace('\\', '/')



import os
import sys
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtCore import QThread
from PyQt5.QtGui import QImage
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QMainWindow
from PyQt5.uic import loadUi

from .camera import CameraDevice

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MainWindow(QMainWindow):

    def __init__(self):
        super().__init__()
        self.ui = loadUi(os.path.join(THIS_DIR, 'mainwindow.ui'), self)

        self.thread = QThread()

        try:
            self.camera = CameraDevice()
        except ValueError:
            self.ui.video.setText("Device not found!\n\nIs FFMPEG available?")
        else:
            self.camera.frame_ready.connect(self.update_video_label)
            self.ui.video.setMinimumSize(*self.camera.size)
            self.camera.moveToThread(self.thread)

    @pyqtSlot(QImage)
    def update_video_label(self, image):
        pixmap = QPixmap.fromImage(image)
        self.ui.video.setPixmap(pixmap)
        self.ui.video.update()


def main():
    app = QApplication(sys.argv)
    main_window = MainWindow()
    main_window.show()
    sys.exit(app.exec_())



import json
import os.path
import sys

def main():
    data_file = os.path.join(os.path.dirname(__file__), 'data.txt')
    with open(data_file, 'r', encoding='utf-8') as f:
        data_from_file = f.read().strip()

    info = {
        'py_executable': sys.executable,
        'py_version': sys.version,
        'data_file_path': data_file,
        'data_file_content': data_from_file,
    }

    print(json.dumps(info, indent=True))



import pytest

from nsist import InputError, split_entry_point

def test_split_entry_point():
    assert split_entry_point('mod:func') == ('mod', 'func')
    assert split_entry_point('mod.submod:func') == ('mod.submod', 'func')

    with pytest.raises(InputError):
        split_entry_point('.\\src\\main:tf2_discord')  # Github issue 205

    with pytest.raises(InputError):
        split_entry_point('mod:1func')  # Identifier can't start with number



import itertools
from operator import itemgetter
import os
import ntpath
import sys

import jinja2

_PKGDIR = os.path.abspath(os.path.dirname(__file__))

PY2 = sys.version_info[0] == 2


class NSISFileWriter(object):
    """Write an .nsi script file by filling in a template.
    """
    def __init__(self, template_file, installerbuilder, definitions=None):
        """Instantiate an NSISFileWriter
        
        :param str template_file: Path to the .nsi template
        :param dict definitions: Mapping of name to value (values will be quoted)
        """
        env = jinja2.Environment(loader=jinja2.FileSystemLoader([
            _PKGDIR,
            os.getcwd()
        ]),
        # Change template markers from {}, which NSIS uses, to [], which it
        # doesn't much, so it's easier to distinguishing our templating from
        # NSIS preprocessor variables.
        block_start_string="[%",
        block_end_string="%]",
        variable_start_string="[[",
        variable_end_string="]]",
        comment_start_string="[#",
        comment_end_string="#]",

        # Trim whitespace around block tags, so formatting works as I expect
        trim_blocks=True,
        lstrip_blocks=True,
        )
        self.template = env.get_template(template_file)
        self.installerbuilder = installerbuilder

        # Group files by their destination directory
        grouped_files = [(dest, [x[0] for x in group]) for (dest, group) in
            itertools.groupby(self.installerbuilder.install_files, itemgetter(1))
                ]
        license_file = None
        if installerbuilder.license_file:
            license_file = os.path.basename(installerbuilder.license_file)
        self.namespace = {
            'ib': installerbuilder,
            'grouped_files': grouped_files,
            'icon': os.path.basename(installerbuilder.icon),
            'arch_tag': '.amd64' if (installerbuilder.py_bitness==64) else '',
            'pjoin': ntpath.join,
            'single_shortcut': len(installerbuilder.shortcuts) == 1,
            'pynsist_pkg_dir': _PKGDIR,
            'has_commands': len(installerbuilder.commands) > 0,
            'python': '"$INSTDIR\\Python\\python"',
            'license_file': license_file,
        }

    def write(self, target):
        """Fill out the template and write the result to 'target'.
        
        :param str target: Path to the file to be written
        """
        with open(target, 'w') as fout:
            fout.write(self.template.render(self.namespace))




import io
import json
from pathlib import Path
import pytest
import re
import responses
from shutil import copytree
from subprocess import run, PIPE
import sys
from testpath import MockCommand, assert_isfile, assert_isdir
from zipfile import ZipFile

from nsist import main
from nsist.util import CACHE_ENV_VAR
from .utils import test_dir, only_on_windows

example_dir = Path(test_dir, 'console_example')

@pytest.fixture()
def console_eg_copy(tmp_path):
    dst = tmp_path / 'example'
    copytree(str(example_dir), str(dst))
    return dst

def respond_python_zip(req):
    buf = io.BytesIO()
    with ZipFile(buf, 'w') as zf:
        zf.writestr('python.exe', b'')
    return 200, {}, buf.getvalue()

@responses.activate
def test_console_example(tmp_path, console_eg_copy, monkeypatch):
    responses.add_callback('GET', re.compile(r'https://www.python.org/ftp/.*'),
        callback=respond_python_zip, content_type='application/zip',
    )

    monkeypatch.chdir(console_eg_copy)
    monkeypatch.setenv(CACHE_ENV_VAR, str(tmp_path / 'cache'))

    with MockCommand('makensis') as makensis:
        ec = main(['installer.cfg'])

    assert ec == 0
    assert makensis.get_calls()[0]['argv'][1].endswith('installer.nsi')

    build_dir = console_eg_copy / 'build' / 'nsis'
    assert_isdir(build_dir)
    assert_isfile(build_dir / 'Python' / 'python.exe')
    assert_isfile(build_dir / 'pkgs' / 'sample_printer' / '__init__.py')
    assert_isfile(build_dir / 'Sample_printer.launch.py')

# To exclude tests requiring network on an unplugged machine, use: pytest -m "not network"

@only_on_windows
@pytest.mark.network
def test_installer(console_eg_copy, tmp_path):
    # Create installer
    run(
        [sys.executable, '-m', 'nsist', 'installer.cfg'],
        cwd=str(console_eg_copy),
        check=True,
    )

    installer = console_eg_copy / 'build' / 'nsis' / 'Sample_printer_1.0.exe'
    inst_dir = tmp_path / 'inst'

    # Run installer
    run([str(installer), '/S', '/INSTDIR={}'.format(inst_dir)], check=True)
    import os
    print(os.listdir(str(inst_dir)))
    inst_python = inst_dir / 'Python' / 'python.exe'
    inst_launch_script = inst_dir / 'Sample_printer.launch.py'
    inst_exe_wrapper = inst_dir / 'bin' / 'pynsist-test-print.exe'

    assert_isfile(inst_python)
    assert_isfile(inst_launch_script)
    assert_isfile(inst_exe_wrapper)
    assert_isfile(inst_dir / 'pkgs' / 'sample_printer' / '__init__.py')


    # Run installed program
    res = run([str(inst_python), str(inst_launch_script)],
              check=True, stdout=PIPE)
    json_res = json.loads(res.stdout.decode('utf-8', 'replace'))
    print(res.stdout.decode('utf-8', 'replace'))

    assert json_res['py_executable'] == str(inst_python)
    assert json_res['py_version'].startswith('3.8.3')  # Set in installer.cfg
    assert json_res['data_file_path'].endswith('data.txt')
    assert json_res['data_file_content'] == 'newt'

    # Run through command-line wrapper
    res2 = run([str(inst_exe_wrapper)], check=True, stdout=PIPE)
    print("Command line wrapper -----------------------------------------")
    print(res2.stdout.decode('utf-8', 'replace'))
    json_res2 = json.loads(res2.stdout.decode('utf-8', 'replace'))

    assert json_res2 == json_res

    # Check command-line wrapper is on PATH
    assert str(inst_exe_wrapper.parent) in get_registry_path()


def get_registry_path():
    """Read the PATH environment variable from the Windows registry"""
    import winreg

    res = []

    # System-wide part of PATH
    with winreg.CreateKey(
        winreg.HKEY_LOCAL_MACHINE,
        r"System\CurrentControlSet\Control\Session Manager\Environment"
    ) as key:
        try:
            path = winreg.QueryValueEx(key, "PATH")[0]
        except WindowsError:
            # No value called PATH
            pass
        else:
            res.extend(path.split(';'))

    # User part of PATH
    with winreg.CreateKey(winreg.HKEY_CURRENT_USER, r"Environment") as key:
        try:
            path = winreg.QueryValueEx(key, "PATH")[0]
        except WindowsError:
            # No value called PATH
            pass
        else:
            res.extend(path.split(';'))

    return res



"""Script to remove build/ directories from all examples.
"""
import os, shutil

examples_dir = os.path.dirname(os.path.abspath(__file__))

for d in os.listdir('.'):
    build_dir = os.path.join(examples_dir, d, 'build')
    if os.path.isdir(build_dir):
        print("Removing {}/build".format(d))
        shutil.rmtree(build_dir)



sys.path.extend([
    os.path.join(pkgdir, 'win32'),
    os.path.join(pkgdir, 'win32', 'lib'),
    os.path.join(pkgdir, 'Pythonwin'),
])

# Preload pywintypes and pythoncom
pwt = os.path.join(pkgdir, 'pywin32_system32', 'pywintypes35.dll')
pcom = os.path.join(pkgdir, 'pywin32_system32', 'pythoncom35.dll')
import imp
imp.load_dynamic('pywintypes', pwt)
imp.load_dynamic('pythoncom', pcom)



from os.path import join as pjoin
from pathlib import Path
import pytest
from testpath import assert_isfile

from nsist.wheels import (
    WheelLocator, extract_wheel, CachedRelease, merge_dir_to, NoWheelError,
    CompatibilityScorer,
)

# To exclude tests requiring network on an unplugged machine, use: pytest -m "not network"

@pytest.mark.network
def test_download(tmpdir):
    tmpdir = str(tmpdir)
    wd = WheelLocator("astsearch==0.1.2", CompatibilityScorer("3.5.1", "win_amd64"))
    wheel = wd.fetch()
    assert_isfile(wheel)

    extract_wheel(wheel, target_dir=tmpdir)
    assert_isfile(pjoin(tmpdir, 'astsearch.py'))
    assert_isfile(pjoin(tmpdir, 'astsearch-0.1.2.dist-info', 'METADATA'))

@pytest.mark.network
def test_bad_name():
    # Packages can't be named after stdlib modules like os
    wl = WheelLocator("os==1.0", "3.5.1", 64)
    with pytest.raises(NoWheelError):
        wl.get_from_pypi()

@pytest.mark.network
def test_bad_version():
    wl = WheelLocator("pynsist==0.99.99", "3.5.1", 64)
    with pytest.raises(NoWheelError):
        wl.get_from_pypi()

def test_extra_sources(tmpdir):
    src1 = Path(str(tmpdir.mkdir('src1')))
    src2 = Path(str(tmpdir.mkdir('src2')))

    # First one found wins, even if a later one is more specific.
    expected = (src1 / 'astsearch-0.1.2-py3-none-any.whl')
    expected.touch()
    (src2 / 'astsearch-0.1.2-py3-none-win_amd64.whl').touch()
    scorer = CompatibilityScorer("3.5.1", "win_amd64")
    wl = WheelLocator("astsearch==0.1.2", scorer, extra_sources=[src1, src2])
    assert wl.check_extra_sources() == expected

    wl = WheelLocator("astsearch==0.2.0", scorer, extra_sources=[src1, src2])
    assert wl.check_extra_sources() is None

def test_pick_best_wheel():
    wd37 = WheelLocator("astsearch==0.1.2", CompatibilityScorer("3.7.1", "win_amd64"))
    wd38 = WheelLocator("astsearch==0.1.2", CompatibilityScorer("3.8.0", "win_amd64"))

    # Some of the wheel filenames below are impossible combinations - they are
    # there to test the scoring and ranking machinery.

    # Prefer Windows-specific wheel
    releases = [
        CachedRelease('astsearch-0.1.2-py3-none-any.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-win_amd64.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[1]

    # Wrong Windows bitness
    releases = [
        CachedRelease('astsearch-0.1.2-py3-none-any.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-win_32.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[0]

    # Prefer more specific Python version
    releases = [
        CachedRelease('astsearch-0.1.2-cp37-none-win_amd64.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-win_amd64.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[0]

    # Prefer more specific Python version
    releases = [
        CachedRelease('astsearch-0.1.2-py34.py37-none-any.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-any.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[0]

    # Incompatible Python version
    releases = [
        CachedRelease('astsearch-0.1.2-cp33-none-any.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-any.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[1]

    # Prefer more specific ABI version
    releases = [
        CachedRelease('astsearch-0.1.2-cp37-abi3-win_amd64.whl'),
        CachedRelease('astsearch-0.1.2-cp37-none-win_amd64.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[0]

    # ABI suffix on Python <3.8
    releases = [
        CachedRelease('astsearch-0.1.2-cp37-cp37-win_amd64.whl'),
        CachedRelease('astsearch-0.1.2-cp37-cp37m-win_amd64.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[1]

    # No ABI suffix on Python >=3.8
    releases = [
        CachedRelease('astsearch-0.1.2-cp38-cp38-win_amd64.whl'),
        CachedRelease('astsearch-0.1.2-cp38-cp38m-win_amd64.whl'),
    ]
    assert wd38.pick_best_wheel(releases) == releases[0]

    # Incompatible ABI version
    releases = [
        CachedRelease('astsearch-0.1.2-cp37-abi4-win_amd64.whl'),
        CachedRelease('astsearch-0.1.2-py3-none-any.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[1]

    # Platform has priority over other attributes
    releases = [
        CachedRelease('astsearch-0.1.2-py37-none-any.whl'),
        CachedRelease('astsearch-0.1.2-py2.py3-none-win_amd64.whl'),
    ]
    assert wd37.pick_best_wheel(releases) == releases[1]

    # Older cp3x tags are compatible when used with abi3
    releases = [
        CachedRelease('cryptography-3.4.7-cp36-cp36m-win_amd64.whl'),
        CachedRelease('cryptography-3.4.7-cp36-abi3-win_amd64.whl'),
    ]
    assert wd38.pick_best_wheel(releases) == releases[1]

def test_merge_dir_to(tmpdir):
    td1 = Path(str(tmpdir.mkdir('one')))
    td2 = Path(str(tmpdir.mkdir('two')))

    with (td1 / 'ab').open('w') as f:
        f.write(u"original")
    with (td2 / 'ab').open('w') as f:
        f.write(u"alternate")

    (td1 / 'subdir').mkdir()
    with (td1 / 'subdir' / 'foo').open('w'): pass
    (td2 / 'subdir').mkdir()
    with (td2 / 'subdir' / 'bar').open('w'): pass

    merge_dir_to(td2, td1)

    assert_isfile(td1 / 'subdir' / 'foo')
    assert_isfile(td1 / 'subdir' / 'bar')
    with (td1 / 'ab').open() as f:
        assert f.read() == u"alternate"

    # Test with conflicts
    (td1 / 'conflict').mkdir()
    with (td2 / 'conflict').open('w'):
        pass

    with pytest.raises(RuntimeError):
        merge_dir_to(td2, td1)
    with pytest.raises(RuntimeError):
        merge_dir_to(td1, td2)



import unittest
import sys

from os.path import dirname

test_dir = dirname(__file__)


def skip_on_windows(function):
    """Decorator to skip a test on Windows."""
    return unittest.skipIf(sys.platform.startswith("win"),
                           "Test for non-Windows platforms")(function)

def only_on_windows(function):
    """Decorator to skip a test on Windows."""
    return unittest.skipUnless(sys.platform.startswith("win"),
                               "Test requires Windows")(function)



#!/usr/bin/python3

import configparser
import os.path
from pathlib import Path

class SectionValidator(object):
    def __init__(self, keys):
        """
        keys
            list of tuples containing the names and whether the
            key is mandatory
        """
        self.keys = keys

    def check(self, config, section_name):
        """
        validates the section, if this is the correct validator for it
        returns True if this is the correct validator for this section

        raises InvalidConfig if something inside the section is wrong
        """
        self._check_mandatory_fields(section_name, config[section_name])
        self._check_invalid_keys(section_name, config[section_name])

    def _check_mandatory_fields(self, section_name, key):
        for key_name, mandatory in self.keys:
            if mandatory:
                try:
                    key[key_name]
                except KeyError:
                    err_msg = ("The section '{0}' must contain a "
                               "key '{1}'!").format(
                                section_name,
                                key_name)
                    raise InvalidConfig(err_msg)

    def _check_invalid_keys(self, section_name, section):
        for key in section:
            key_name = str(key)
            valid_key_names = [s[0] for s in self.keys]
            is_valid_key = key_name in valid_key_names
            if not is_valid_key:
                err_msg = ("'{0}' is not a valid key name for '{1}'. Must "
                           "be one of these: {2}").format(
                            key_name,
                            section_name,
                            ', '.join(valid_key_names))
                raise InvalidConfig(err_msg)

# contains all configuration sections and keys
# the keys are a tuple with their name and a boolean, which
# tells us whether the option is mandatory
CONFIG_VALIDATORS = {
    'Application': SectionValidator([
        ('name', True),
        ('version', True),
        ('publisher', False),
        ('entry_point', False),
        ('script', False),
        ('target', False),
        ('parameters', False),
        ('icon', False),
        ('console', False),
        ('extra_preamble', False),
        ('license_file', False),
    ]),
    'Build': SectionValidator([
        ('directory', False),
        ('installer_name', False),
        ('nsi_template', False),
    ]),
    'Include': SectionValidator([
        ('packages', False),
        ('pypi_wheels', False),
        ('extra_wheel_sources', False),
        ('files', False),
        ('exclude', False),
        ('local_wheels', False)
    ]),
    'Python': SectionValidator([
        ('version', False),
        ('bitness', False),
        ('format', False),
        ('include_msvcrt', False),
    ]),
    'Shortcut': SectionValidator([
        ('entry_point', False),
        ('script', False),
        ('target', False),
        ('parameters', False),
        ('icon', False),
        ('console', False),
        ('extra_preamble', False),
    ]),
    'Command': SectionValidator([
        ('entry_point', True),
        ('console', False),
        ('extra_preamble', False),
    ])
}

class InvalidConfig(ValueError):
    pass

def read_and_validate(config_file):
    # Interpolation interferes with Windows-style environment variables, so
    # it's disabled for now.
    config = configparser.ConfigParser(interpolation=None)
    if config.read(config_file) == []:
        raise InvalidConfig("Config file not found: %r" % config_file)
    for section in config.sections():
        if section in CONFIG_VALIDATORS:
            CONFIG_VALIDATORS[section].check(config, section)
        elif section.startswith('Shortcut '):
            CONFIG_VALIDATORS['Shortcut'].check(config, section)
        elif section.startswith('Command '):
            CONFIG_VALIDATORS['Command'].check(config, section)
        else:
            valid_section_names = CONFIG_VALIDATORS.keys()
            err_msg = ("{0} is not a valid section header. Must "
                       "be one of these: {1}").format(
                       section,
                       ', '.join(['"%s"' % n for n in valid_section_names]))
            raise InvalidConfig(err_msg)
    return config

def read_extra_files(cfg):
    """Read the list of extra files from the config file.

    Returns a list of 2-tuples: (file, destination_directory), which can be
    passed as the ``extra_files`` parameter to :class:`nsist.InstallerBuilder`.
    """
    lines = cfg.get('Include', 'files', fallback='').splitlines()
    pairs = []
    for line in lines:
        if '>' in line:
            file, dest = line.rsplit('>', 1)
            pairs.append((file.strip(), dest.strip()))
        else:
            pairs.append((line, '$INSTDIR'))

    return pairs

def read_shortcuts_config(cfg):
    """Read and verify the shortcut definitions from the config file.

    There is one shortcut per 'Shortcut <name>' section, and one for the
    Application section.

    Returns a dict of dicts with the fields from the shortcut sections.
    The optional 'icon' and 'console' fields will be filled with their
    default values if not supplied.
    """
    shortcuts = {}
    def _check_shortcut(name, sc, section):
        alternatives = ['entry_point', 'script', 'target']
        has_alternatives = sum(1 for k in alternatives if k in sc)
        if has_alternatives < 1:
            raise InvalidConfig('Section [{}] has none of {}.'.format(
                                    section, ', '.join(alternatives)))
        elif has_alternatives > 1:
            raise InvalidConfig('Section [{}] has more than one of {}.'.format(
                                    section, ', '.join(alternatives)))

        # Copy to a regular dict so it can hold a boolean value
        sc2 = dict(sc)
        if 'icon' not in sc2:
            from . import DEFAULT_ICON
            sc2['icon'] = DEFAULT_ICON
        sc2['console'] = sc.getboolean('console', fallback=False)
        sc2['parameters'] = sc.get('parameters', fallback='')
        if 'extra_preamble' in sc2:
            if 'entry_point' not in sc2:
                raise InvalidConfig('extra_preamble is only valid with entry_point')
            preamb_file = sc2['extra_preamble']
            if not os.path.isfile(preamb_file):
                raise InvalidConfig('extra_preamble file %r does not exist' %
                                                    preamb_file)

        shortcuts[name] = sc2

    for section in cfg.sections():
        if section.startswith("Shortcut "):
            name = section[len("Shortcut "):]
            _check_shortcut(name, cfg[section], section)

    appcfg = cfg['Application']
    _check_shortcut(appcfg['name'], appcfg, 'Application')

    return shortcuts

def read_commands_config(cfg):
    """Read and verify the command definitions from the config file.

    Returns a dict of dicts, keyed by command name, containing the values from
    the command sections of the config file.
    """
    commands = {}
    for section in cfg.sections():
        if section.startswith("Command "):
            name = section[len("Command "):]
            commands[name] = cc = dict(cfg[section])
            cc['console'] = cfg[section].getboolean('console', fallback=True)
            if ('extra_preamble' in cc) and \
                    not os.path.isfile(cc['extra_preamble']):
                raise InvalidConfig('extra_preamble file %r does not exist' %
                                    cc['extra_preamble'])

    return commands


def get_installer_builder_args(config):
    from . import (DEFAULT_BITNESS,
                    DEFAULT_BUILD_DIR,
                    DEFAULT_ICON,
                    DEFAULT_PY_VERSION)

    appcfg = config['Application']
    args = {}
    args['appname'] = appcfg['name']
    args['version'] = appcfg['version']
    args['shortcuts'] = read_shortcuts_config(config)
    args['commands'] = read_commands_config(config)
    args['publisher'] = appcfg.get('publisher', None)
    args['icon'] = appcfg.get('icon', DEFAULT_ICON)
    args['license_file'] = appcfg.get('license_file', None)
    args['packages'] = config.get('Include', 'packages', fallback='').strip().splitlines()
    args['pypi_wheel_reqs'] = config.get('Include', 'pypi_wheels', fallback='').strip().splitlines()
    args['extra_wheel_sources'] = [Path(p) for p in
        config.get('Include', 'extra_wheel_sources', fallback='').strip().splitlines()
    ]
    args['extra_files'] = read_extra_files(config)
    args['py_version'] = config.get('Python', 'version', fallback=DEFAULT_PY_VERSION)
    args['py_bitness'] = config.getint('Python', 'bitness', fallback=DEFAULT_BITNESS)
    args['inc_msvcrt'] = config.getboolean('Python', 'include_msvcrt', fallback=True)
    args['build_dir'] = config.get('Build', 'directory', fallback=DEFAULT_BUILD_DIR)
    args['installer_name'] = config.get('Build', 'installer_name', fallback=None)
    args['nsi_template'] = config.get('Build', 'nsi_template', fallback=None)
    args['exclude'] = config.get('Include', 'exclude', fallback='').strip().splitlines()
    args['local_wheels'] = config.get('Include', 'local_wheels', fallback='').strip().splitlines()
    return args



"""Build NSIS installers for Python applications.
"""
import io
import logging
import ntpath
import operator
import os
from pathlib import Path
import re
import shutil
from subprocess import call
import sys
import fnmatch
import zipfile

if os.name == 'nt':
    import winreg
else:
    winreg = None

from .configreader import get_installer_builder_args
from .commands import prepare_bin_directory
from .copymodules import copy_modules
from .nsiswriter import NSISFileWriter
from .wheels import WheelGetter
from .util import download, get_cache_dir, normalize_path

__version__ = '2.8'

pjoin = os.path.join
logger = logging.getLogger(__name__)

_PKGDIR = os.path.abspath(os.path.dirname(__file__))
DEFAULT_PY_VERSION = '3.6.3'
DEFAULT_BUILD_DIR = pjoin('build', 'nsis')
DEFAULT_ICON = pjoin(_PKGDIR, 'glossyorb.ico')
if os.name == 'nt' and sys.maxsize == (2**63)-1:
    DEFAULT_BITNESS = 64
else:
    DEFAULT_BITNESS = 32

def find_makensis_win():
    """Locate makensis.exe on Windows by querying the registry"""
    try:
        nsis_install_dir = winreg.QueryValue(winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\NSIS')
    except OSError:
        nsis_install_dir = winreg.QueryValue(winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\Wow6432Node\\NSIS')

    return pjoin(nsis_install_dir, 'makensis.exe')

class InputError(ValueError):
    def __init__(self, param, value, expected):
        self.param = param
        self.value = value
        self.expected = expected

    def __str__(self):
        return "{e.value!r} is not valid for {e.param}, expected {e.expected}".format(e=self)


def split_entry_point(ep: str):
    """Like ep.split(':'), but with extra checks and helpful errors"""
    module, _, func = ep.partition(':')
    if all([s.isidentifier() for s in module.split('.')]) and func.isidentifier():
        return module, func

    raise InputError(
        'entry point', ep, "'mod:func', so 'from mod import func' works"
    )


class InstallerBuilder(object):
    """Controls building an installer. This includes three main steps:

    1. Arranging the necessary files in the build directory.
    2. Filling out the template NSI file to control NSIS.
    3. Running ``makensis`` to build the installer.

    :param str appname: Application name
    :param str version: Application version
    :param dict shortcuts: Dictionary keyed by shortcut name, containing
            dictionaries whose keys match the fields of :ref:`shortcut_config`
            in the config file
    :param str publisher: Publisher name
    :param str icon: Path to an icon for the application
    :param list packages: List of strings for importable packages to include
    :param dict commands: Dictionary keyed by command name, containing dicts
            defining the commands, as in the config file.
    :param list pypi_wheel_reqs: Package specifications to fetch from PyPI as wheels
    :param extra_wheel_sources: Directory paths to find wheels in.
    :type extra_wheel_sources: list of Path objects
    :param local_wheels: Glob paths matching wheel files to include
    :type local_wheels: list of str
    :param list extra_files: List of 2-tuples (file, destination) of files to include
    :param list exclude: Paths of files to exclude that would otherwise be included
    :param str py_version: Full version of Python to bundle
    :param int py_bitness: Bitness of bundled Python (32 or 64)
    :param str py_format: (deprecated) 'bundled'. Use Pynsist 1.x for
            'installer' option.
    :param bool inc_msvcrt: True to include the Microsoft C runtime with 'bundled'
            Python.
    :param str build_dir: Directory to run the build in
    :param str installer_name: Filename of the installer to produce
    :param str nsi_template: Path to a template NSI file to use
    """
    def __init__(self, appname, version, shortcuts, *, publisher=None,
                icon=DEFAULT_ICON, packages=None, extra_files=None,
                py_version=DEFAULT_PY_VERSION, py_bitness=DEFAULT_BITNESS,
                py_format='bundled', inc_msvcrt=True, build_dir=DEFAULT_BUILD_DIR,
                installer_name=None, nsi_template=None,
                exclude=None, pypi_wheel_reqs=None, extra_wheel_sources=None,
                local_wheels=None, commands=None, license_file=None):
        self.appname = appname
        self.version = version
        self.publisher = publisher
        self.shortcuts = shortcuts
        self.icon = icon
        self.packages = packages or []
        self.exclude = [normalize_path(p) for p in (exclude or [])]
        self.extra_files = extra_files or []
        self.pypi_wheel_reqs = pypi_wheel_reqs or []
        self.extra_wheel_sources = extra_wheel_sources or []
        self.local_wheels = local_wheels or []
        self.commands = commands or {}
        self.license_file = license_file

        # Python options
        self.py_version = py_version
        if not self._py_version_pattern.match(py_version):
            if not os.environ.get('PYNSIST_PY_PRERELEASE'):
                raise InputError('py_version', py_version,
                                 "a full Python version like '3.4.0'")
        if self.py_version_tuple < (3, 5):
            raise InputError('py_version', py_version,
                             "Python >= 3.5.0 (use Pynsist 1.x for older Python.")
        self.py_bitness = py_bitness
        if py_bitness not in {32, 64}:
            raise InputError('py_bitness', py_bitness, "32 or 64")
        self.py_major_version = self.py_qualifier = '.'.join(self.py_version.split('.')[:2])
        if self.py_bitness == 32:
            self.py_qualifier += '-32'

        if py_format == 'installer':
            raise InputError('py_format', py_format, "'bundled' (use Pynsist 1.x for 'installer')")
        elif py_format != 'bundled':
            raise InputError('py_format', py_format, "'bundled'")

        self.inc_msvcrt = inc_msvcrt

        # Build details
        self.build_dir = build_dir
        self.installer_name = installer_name or self.make_installer_name()
        self.nsi_template = nsi_template
        if self.nsi_template is None:
            if self.inc_msvcrt:
                self.nsi_template = 'pyapp_msvcrt.nsi'
            else:
                self.nsi_template = 'pyapp.nsi'

        self.nsi_file = pjoin(self.build_dir, 'installer.nsi')

        # To be filled later
        self.install_files = []
        self.install_dirs = []
        self.msvcrt_files = []

    _py_version_pattern = re.compile(r'\d\.\d+\.\d+$')

    @property
    def py_version_tuple(self):
        parts = self.py_version.split('.')
        return int(parts[0]), int(parts[1])

    def make_installer_name(self):
        """Generate the filename of the installer exe

        e.g. My_App_1.0.exe
        """
        s = self.appname + '_' + self.version + '.exe'
        return s.replace(' ', '_')

    def _python_download_url_filename(self):
        version = self.py_version
        bitness = self.py_bitness
        filename = 'python-{}-embed-{}.zip'.format(version,
                                   'amd64' if bitness==64 else 'win32')

        version_minus_prerelease = re.sub(r'(a|b|rc)\d+$', '', self.py_version)
        return 'https://www.python.org/ftp/python/{0}/{1}'.format(
                version_minus_prerelease, filename), filename

    def fetch_python_embeddable(self):
        """Fetch the embeddable Windows build for the specified Python version

        It will be unpacked into the build directory.

        In addition, any ``*._pth`` files found therein will have the pkgs path
        appended to them.
        """
        url, filename = self._python_download_url_filename()
        cache_file = get_cache_dir(ensure_existence=True) / filename
        if not cache_file.is_file():
            logger.info('Downloading embeddable Python build...')
            logger.info('Getting %s', url)
            download(url, cache_file)

        logger.info('Unpacking Python...')
        python_dir = pjoin(self.build_dir, 'Python')

        with zipfile.ZipFile(str(cache_file)) as z:
            z.extractall(python_dir)

        # Manipulate any *._pth files so the default paths AND pkgs directory
        # ends up in sys.path. Please see:
        # https://docs.python.org/3/using/windows.html#finding-modules
        # for more information.
        pth_files = [f for f in os.listdir(python_dir)
                     if os.path.isfile(pjoin(python_dir, f))
                     and f.endswith('._pth')]
        for pth in pth_files:
            with open(pjoin(python_dir, pth), 'a+b') as f:
                f.write(b'\r\n..\\pkgs\r\nimport site\r\n')

        self.install_dirs.append(('Python', '$INSTDIR'))

    def prepare_msvcrt(self):
        arch = 'x64' if self.py_bitness == 64 else 'x86'
        src = pjoin(_PKGDIR, 'msvcrt', arch)
        dst = pjoin(self.build_dir, 'msvcrt')
        self.msvcrt_files = sorted(os.listdir(src))

        shutil.copytree(src, dst)

    SCRIPT_TEMPLATE = """#!python{qualifier}
import sys, os
import site
scriptdir, script = os.path.split(os.path.abspath(__file__))
installdir = scriptdir  # for compatibility with commands
pkgdir = os.path.join(scriptdir, 'pkgs')
sys.path.insert(0, pkgdir)
# Ensure .pth files in pkgdir are handled properly
site.addsitedir(pkgdir)
os.environ['PYTHONPATH'] = pkgdir + os.pathsep + os.environ.get('PYTHONPATH', '')

# APPDATA should always be set, but in case it isn't, try user home
# If none of APPDATA, HOME, USERPROFILE or HOMEPATH are set, this will fail.
appdata = os.environ.get('APPDATA', None) or os.path.expanduser('~')

if 'pythonw' in sys.executable:
    # Running with no console - send all stdstream output to a file.
    kw = {{'errors': 'replace'}} if (sys.version_info[0] >= 3) else {{}}
    sys.stdout = sys.stderr = open(os.path.join(appdata, script+'.log'), 'w', **kw)
else:
    # In a console. But if the console was started just for this program, it
    # will close as soon as we exit, so write the traceback to a file as well.
    def excepthook(etype, value, tb):
        "Write unhandled exceptions to a file and to stderr."
        import traceback
        traceback.print_exception(etype, value, tb)
        with open(os.path.join(appdata, script+'.log'), 'w') as f:
            traceback.print_exception(etype, value, tb, file=f)
    sys.excepthook = excepthook

{extra_preamble}

if __name__ == '__main__':
    from {module} import {func}
    {func}()
"""

    def write_script(self, entrypt, target, extra_preamble=''):
        """Write a launcher script from a 'module:function' entry point

        py_version and py_bitness are used to write an appropriate shebang line
        for the PEP 397 Windows launcher.
        """
        module, func = split_entry_point(entrypt)
        with open(target, 'w') as f:
            f.write(self.SCRIPT_TEMPLATE.format(qualifier=self.py_qualifier,
                    module=module, func=func, extra_preamble=extra_preamble))

        pkg = module.split('.')[0]
        if pkg not in self.packages:
            self.packages.append(pkg)

    def prepare_shortcuts(self):
        """Prepare shortcut files in the build directory.

        If entry_point is specified, write the script. If script is specified,
        copy to the build directory. Prepare target and parameters for these
        shortcuts.

        Also copies shortcut icons.
        """
        files = set()
        for scname, sc in self.shortcuts.items():
            if not sc.get('target'):
                if sc.get('entry_point'):
                    sc['script'] = script = scname.replace(' ', '_') + '.launch.py' \
                                                + ('' if sc['console'] else 'w')

                    specified_preamble = sc.get('extra_preamble', None)
                    if isinstance(specified_preamble, str):
                        # Filename
                        extra_preamble = io.open(specified_preamble, encoding='utf-8')
                    elif specified_preamble is None:
                        extra_preamble = io.StringIO()  # Empty
                    else:
                        # Passed a StringIO or similar object
                        extra_preamble = specified_preamble

                    self.write_script(sc['entry_point'], pjoin(self.build_dir, script),
                                      extra_preamble.read().rstrip())
                else:
                    shutil.copy2(sc['script'], self.build_dir)

                target = '$INSTDIR\\Python\\python{}.exe'
                sc['target'] = target.format('' if sc['console'] else 'w')
                sc['parameters'] = '"%s"' % ntpath.join('$INSTDIR', sc['script'])
                files.add(os.path.basename(sc['script']))

            shutil.copy2(sc['icon'], self.build_dir)
            sc['icon'] = os.path.basename(sc['icon'])
            files.add(sc['icon'])
        self.install_files.extend([(f, '$INSTDIR') for f in files])

    def copy_license(self):
        """
        If a license file has been specified, ensure it's copied into the
        install directory and added to the install_files list.
        """
        if self.license_file:
            shutil.copy2(self.license_file, self.build_dir)
            license_file_name = os.path.basename(self.license_file)
            self.install_files.append((license_file_name, '$INSTDIR'))


    def prepare_packages(self):
        """Move requested packages into the build directory.

        If a pynsist_pkgs directory exists, it is copied into the build
        directory as pkgs/ . Any packages not already there are found on
        sys.path and copied in.
        """
        logger.info("Copying packages into build directory...")
        build_pkg_dir = pjoin(self.build_dir, 'pkgs')

        # 1. Manually prepared packages
        if os.path.isdir('pynsist_pkgs'):
            shutil.copytree('pynsist_pkgs', build_pkg_dir)
        else:
            os.mkdir(build_pkg_dir)

        # 2. Wheels specified in pypi_wheel_reqs or in paths of local_wheels
        wg = WheelGetter(self.pypi_wheel_reqs, self.local_wheels, build_pkg_dir,
                         py_version=self.py_version, bitness=self.py_bitness,
                         extra_sources=self.extra_wheel_sources,
                         exclude=self.exclude)
        wg.get_all()

        # 3. Copy importable modules
        copy_modules(self.packages, build_pkg_dir,
                     py_version=self.py_version, exclude=self.exclude)

    def prepare_commands(self):
        for cmd in self.commands.values():
            split_entry_point(cmd['entry_point'])  # Check entry point format
        command_dir = Path(self.build_dir) / 'bin'
        command_dir.mkdir()
        prepare_bin_directory(command_dir, self.commands, bitness=self.py_bitness)
        self.install_dirs.append((command_dir.name, '$INSTDIR'))
        self.extra_files.append((pjoin(_PKGDIR, '_system_path.py'), '$INSTDIR'))

    def copytree_ignore_callback(self, directory, files):
        """This is being called back by our shutil.copytree call to implement the
        'exclude' feature.
        """
        ignored = set()

        # Filter by file names relative to the build directory
        directory = os.path.normpath(directory)
        files = [os.path.join(directory, fname) for fname in files]

        # Execute all patterns
        for pattern in self.exclude:
            ignored.update([
                os.path.basename(fname)
                for fname in fnmatch.filter(files, pattern)
            ])
        return ignored

    def copy_extra_files(self):
        """Copy a list of files into the build directory, and add them to
        install_files or install_dirs as appropriate.
        """
        # Create installer.nsi, so that a data file with the same name will
        # automatically be renamed installer.1.nsi. All the other files needed
        # in the build directory should already be in place.
        Path(self.nsi_file).touch()

        for file, destination in self.extra_files:
            file = file.rstrip('/\\')
            in_build_dir = Path(self.build_dir, os.path.basename(file))

            # Find an unused name in the build directory,
            # similar to the source filename, e.g. foo.1.txt, foo.2.txt, ...
            stem, suffix = in_build_dir.stem, in_build_dir.suffix
            n = 1
            while in_build_dir.exists():
                name = '{}.{}{}'.format(stem, n, suffix)
                in_build_dir = in_build_dir.with_name(name)
                n += 1

            if destination:
                # Normalize destination paths to Windows-style
                destination = destination.replace('/', '\\')
            else:
                destination = '$INSTDIR'

            if os.path.isdir(file):
                if self.exclude:
                    shutil.copytree(file, str(in_build_dir),
                                    ignore=self.copytree_ignore_callback)
                else:
                    # Don't use our exclude callback if we don't need to,
                    # as it slows things down.
                    shutil.copytree(file, str(in_build_dir))
                self.install_dirs.append((in_build_dir.name, destination))
            else:
                shutil.copy2(file, str(in_build_dir))
                self.install_files.append((in_build_dir.name, destination))

    def write_nsi(self):
        """Write the NSI file to define the NSIS installer.

        Most of the details of this are in the template and the
        :class:`nsist.nsiswriter.NSISFileWriter` class.
        """
        nsis_writer = NSISFileWriter(self.nsi_template, installerbuilder=self)

        logger.info('Writing NSI file to %s', self.nsi_file)
        # Sort by destination directory, so we can group them effectively
        self.install_files.sort(key=operator.itemgetter(1))
        nsis_writer.write(self.nsi_file)

    def run_nsis(self):
        """Runs makensis using the specified .nsi file

        Returns the exit code.
        """
        makensis = shutil.which('makensis')
        if (makensis is None) and os.name == 'nt':
            try:
                makensis = find_makensis_win()
            except OSError:
                pass

        if makensis is None:
            print("makensis was not found. Install NSIS and try again.")
            print("http://nsis.sourceforge.net/Download")
            return 1

        logger.info('\n~~~ Running makensis ~~~')
        return call([makensis, self.nsi_file])

    def run(self, makensis=True):
        """Run all the steps to build an installer.
        """
        try:  # Start with a clean build directory
            shutil.rmtree(self.build_dir)
        except FileNotFoundError:
            pass
        os.makedirs(self.build_dir)

        self.fetch_python_embeddable()
        if self.inc_msvcrt:
            self.prepare_msvcrt()

        self.prepare_shortcuts()

        self.copy_license()

        if self.commands:
            self.prepare_commands()

        # Packages
        self.prepare_packages()

        # Extra files
        self.copy_extra_files()

        self.write_nsi()

        if makensis:
            exitcode = self.run_nsis()

            if not exitcode:
                logger.info('Installer written to %s', pjoin(self.build_dir, self.installer_name))
            return exitcode
        return 0

def main(argv=None):
    """Make an installer from the command line.

    This parses command line arguments and a config file, and calls
    :func:`all_steps` with the extracted information.
    """
    logger.setLevel(logging.INFO)
    logger.handlers = [logging.StreamHandler()]

    import argparse
    argp = argparse.ArgumentParser(prog='pynsist')
    argp.add_argument('config_file')
    argp.add_argument('--no-makensis', action='store_true',
        help='Prepare files and folders, stop before calling makensis. For debugging.'
    )
    options = argp.parse_args(argv)

    dirname, config_file = os.path.split(options.config_file)
    if dirname:
        os.chdir(dirname)

    from . import configreader
    try:
        cfg = configreader.read_and_validate(config_file)
        args = get_installer_builder_args(cfg)
    except configreader.InvalidConfig as e:
        logger.error('Error parsing configuration file:')
        logger.error(str(e))
        sys.exit(1)

    try:
        ec = InstallerBuilder(**args).run(makensis=(not options.no_makensis))
    except InputError as e:
        logger.error("Error in config values:")
        logger.error(str(e))
        sys.exit(1)

    return ec



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(393, 606)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radio_1 = QtWidgets.QRadioButton(self.groupBox)
        self.radio_1.setObjectName("radio_1")
        self.verticalLayout.addWidget(self.radio_1)
        self.radio_2 = QtWidgets.QRadioButton(self.groupBox)
        self.radio_2.setObjectName("radio_2")
        self.verticalLayout.addWidget(self.radio_2)
        self.radio_3 = QtWidgets.QRadioButton(self.groupBox)
        self.radio_3.setObjectName("radio_3")
        self.verticalLayout.addWidget(self.radio_3)
        self.radio_4 = QtWidgets.QRadioButton(self.groupBox)
        self.radio_4.setObjectName("radio_4")
        self.verticalLayout.addWidget(self.radio_4)
        self.add_button = QtWidgets.QPushButton(self.groupBox)
        self.add_button.setObjectName("add_button")
        self.verticalLayout.addWidget(self.add_button)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout.addWidget(self.groupBox)
        self.listWidget = QtWidgets.QListWidget(self.centralwidget)
        self.listWidget.setObjectName("listWidget")
        self.horizontalLayout.addWidget(self.listWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 393, 24))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionNew = QtWidgets.QAction(MainWindow)
        self.actionNew.setObjectName("actionNew")
        self.actionOpen = QtWidgets.QAction(MainWindow)
        self.actionOpen.setObjectName("actionOpen")
        self.actionSave = QtWidgets.QAction(MainWindow)
        self.actionSave.setObjectName("actionSave")
        self.menuFile.addAction(self.actionNew)
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionSave)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.groupBox.setTitle(_translate("MainWindow", "Click things"))
        self.radio_1.setText(_translate("MainWindow", "Thing 1"))
        self.radio_2.setText(_translate("MainWindow", "Thing 2"))
        self.radio_3.setText(_translate("MainWindow", "Thing 3"))
        self.radio_4.setText(_translate("MainWindow", "Last thing"))
        self.add_button.setText(_translate("MainWindow", "Add to list"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionNew.setText(_translate("MainWindow", "New"))
        self.actionOpen.setText(_translate("MainWindow", "Open"))
        self.actionSave.setText(_translate("MainWindow", "Save"))




import importlib
import importlib.abc
import importlib.machinery
import os
import shutil
import sys
import tempfile
import zipfile, zipimport
import fnmatch
from functools import partial

from .util import normalize_path

pjoin = os.path.join

PY2 = sys.version_info[0] == 2
running_python  = '.'.join(str(x) for x in sys.version_info[:2])

class ExtensionModuleMismatch(ImportError):
    pass

extensionmod_errmsg = """Found an extension module that will not be usable on %s:
%s
Put Windows packages in pynsist_pkgs/ to avoid this."""

def check_ext_mod(path, target_python):
    """If path is an extension module, check that it matches target platform.
    
    It should be for Windows and we should be running on the same version
    of Python that we're targeting. Raises ExtensionModuleMismatch if not.
    
    Does nothing if path is not an extension module.
    """
    if path.endswith('.so'):
        raise ExtensionModuleMismatch(extensionmod_errmsg % ('Windows', path))
    elif path.endswith('.pyd') and not target_python.startswith(running_python):
        # TODO: From Python 3.2, extension modules can restrict themselves
        # to a stable ABI. Can we detect this?
        raise ExtensionModuleMismatch(extensionmod_errmsg % ('Python '+target_python, path))

def check_package_for_ext_mods(path, target_python):
    """Walk the directory path, calling :func:`check_ext_mod` on each file.
    """
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            check_ext_mod(os.path.join(path, dirpath, filename), target_python)

def copy_zipmodule(loader, modname, target):
    """Copy a module or package out of a zip file to the target directory."""
    file = loader.get_filename(modname)
    assert file.startswith(loader.archive)
    path_in_zip = file[len(loader.archive+'/'):]
    zf = zipfile.ZipFile(loader.archive)

    # If the packages are in a subdirectory, extracting them recreates the
    # directory structure from the zip file. So extract to a temp dir first,
    # and then copy the modules to target.
    tempdir = tempfile.mkdtemp()
    if loader.is_package(modname):
        # Extract everything in a folder
        pkgdir, basename = os.path.split(path_in_zip)
        assert basename.startswith('__init__')
        pkgfiles = [f for f in zf.namelist() if f.startswith(pkgdir)]
        zf.extractall(tempdir, pkgfiles)
        shutil.copytree(pjoin(tempdir, pkgdir), pjoin(target, modname))
    else:
        # Extract a single file
        zf.extract(path_in_zip, tempdir)
        shutil.copy2(pjoin(tempdir, path_in_zip), target)

    shutil.rmtree(tempdir)

def copytree_ignore_callback(excludes, pkgdir, modname, directory, files):
    """This is being called back by our shutil.copytree call to implement the
    'exclude' feature.
    """
    ignored = set()

    # Filter by file names relative to the build directory
    reldir = os.path.relpath(directory, pkgdir)
    target = os.path.join('pkgs', modname, reldir)
    files = [normalize_path(os.path.join(target, fname)) for fname in files]
    # Execute all patterns
    for pattern in excludes + ['*.pyc']:
        ignored.update([
            os.path.basename(fname)
            for fname in fnmatch.filter(files, pattern)
        ])

    return ignored


class ModuleCopier:
    """Finds and copies importable Python modules and packages.

    This is the Python >3.3 version and uses the `importlib` package to
    locate modules.
    """
    def __init__(self, py_version, path=None):
        self.py_version = py_version
        self.path = path if (path is not None) else ([''] + sys.path)

    def copy(self, modname, target, exclude):
        """Copy the importable module 'modname' to the directory 'target'.

        modname should be a top-level import, i.e. without any dots.
        Packages are always copied whole.

        This can currently copy regular filesystem files and directories,
        and extract modules and packages from appropriately structured zip
        files.
        """
        spec = importlib.machinery.PathFinder.find_spec(modname, self.path)
        if spec is None:
            raise ImportError('Could not find %r' % modname)
        loader = spec.loader
        if loader is None:
            raise ImportError('Cannot bundle namespace package %r' % modname)

        pkg = loader.is_package(modname)

        if isinstance(loader, importlib.machinery.ExtensionFileLoader):
            check_ext_mod(loader.path, self.py_version)
            shutil.copy2(loader.path, target)

        elif isinstance(loader, importlib.abc.FileLoader):
            file = loader.get_filename(modname)
            if pkg:
                pkgdir, basename = os.path.split(file)
                assert basename.startswith('__init__')
                check_package_for_ext_mods(pkgdir, self.py_version)
                dest = os.path.join(target, modname)
                if exclude:
                    shutil.copytree(
                        pkgdir, dest,
                        ignore=partial(copytree_ignore_callback, exclude, pkgdir, modname)
                    )
                else:
                    # Don't use our exclude callback if we don't need to,
                    # as it slows things down.
                    shutil.copytree(
                        pkgdir, dest,
                        ignore=shutil.ignore_patterns('*.pyc')
                    )
            else:
                shutil.copy2(file, target)

        elif isinstance(loader, zipimport.zipimporter):
            copy_zipmodule(loader, modname, target)


def copy_modules(modnames, target, py_version, path=None, exclude=None):
    """Copy the specified importable modules to the target directory.

    By default, it finds modules in :data:`sys.path` - this can be overridden
    by passing the path parameter.
    """
    mc = ModuleCopier(py_version, path)
    files_in_target_noext = [os.path.splitext(f)[0] for f in os.listdir(target)]

    for modname in modnames:
        if modname in files_in_target_noext:
            # Already there, no need to copy it.
            continue
        mc.copy(modname, target, exclude)

    if not modnames:
        # NSIS abhors an empty folder, so give it a file to find.
        with open(os.path.join(target, 'placeholder'), 'w') as f:
            f.write('This file only exists so NSIS finds something in this directory.')



import configparser
import os

import pytest

from .. import configreader


DATA_FILES = os.path.join(os.path.dirname(__file__), 'data_files')

def test_valid_config():
    configfile = os.path.join(DATA_FILES, 'valid_config.cfg')
    config = configreader.read_and_validate(configfile)
    assert config.has_section('Application')
    args = configreader.get_installer_builder_args(config)
    assert args['py_version'] == '3.4.0'

def test_valid_config_with_shortcut():
    configfile = os.path.join(DATA_FILES, 'valid_config_with_shortcut.cfg')
    configreader.read_and_validate(configfile)

def test_valid_config_with_commands():
    configfile = os.path.join(DATA_FILES, 'valid_config_with_commands.cfg')
    configreader.read_and_validate(configfile)

def test_valid_config_with_values_starting_on_new_line():
    configfile = os.path.join(DATA_FILES, 'valid_config_value_newline.cfg')
    config = configreader.read_and_validate(configfile)
    sections = ('Application', 'Python', 'Include', 'Build')
    assert len(config.sections()) == len(sections)
    for section in sections:
        assert section in config

    assert config.get('Include', 'packages') == '\nrequests\nbs4'
    assert config.get('Include', 'pypi_wheels') == '\nhtml5lib'
    assert config.get('Include', 'exclude') == '\nsomething'
    assert config.get('Include', 'files') == '\nLICENSE\ndata_files/'

    args = configreader.get_installer_builder_args(config)
    assert args['appname'] == 'My App'
    assert args['version'] == '1.0'
    assert args['shortcuts']['My App']['entry_point'] == 'myapp:main'
    assert args['commands'] == {}
    assert args['publisher'] == 'Test'
    # assert args['entry_point'] == '\nmyapp:main'
    assert args['icon'] == 'myapp.ico'

    assert args['py_version'] == '3.6.0'
    assert args['py_bitness'] == 64
    assert args['inc_msvcrt'] is True

    assert args['build_dir'] == 'build/'
    assert args['nsi_template'] == 'template.nsi'

    assert args['packages'] == ['requests', 'bs4']
    assert args['pypi_wheel_reqs'] == ['html5lib']
    assert args['exclude'] == ['something']
    assert args['extra_files'] == [('', '$INSTDIR'),
                                   ('LICENSE', '$INSTDIR'),
                                   ('data_files/', '$INSTDIR')]

def test_invalid_config_keys():
    with pytest.raises(configreader.InvalidConfig):
        configfile = os.path.join(DATA_FILES, 'invalid_config_section.cfg')
        configreader.read_and_validate(configfile)

def test_invalid_config_key():
    with pytest.raises(configreader.InvalidConfig):
        configfile = os.path.join(DATA_FILES, 'invalid_config_key.cfg')
        configreader.read_and_validate(configfile)

def test_missing_config_key():
    with pytest.raises(configreader.InvalidConfig):
        configfile = os.path.join(DATA_FILES, 'missing_config_key.cfg')
        configreader.read_and_validate(configfile)

def test_invalid_config_file():
    with pytest.raises(configparser.Error):
        configfile = os.path.join(DATA_FILES, 'not_a_config.cfg')
        configreader.read_and_validate(configfile)



"""A fun number guessing game!"""

import random

def main():
    number = random.randint(1, 100)
    guesses = 0

    print("I'm thinking of a number, between 1 and 100. Can you guess what it is?")
    while True:
        guesses += 1
        guess = input('= ')
        try:
            guess = int(guess)
        except ValueError:
            print("Base 10 integers only, please.")
            continue

        if guess > number:
            print("Too high!")
        elif guess <  number:
            print("Too low!")
        else:
            print("That's right, {}. You got it in {} guesses.".format(number, guesses))
            break

    print()
    input("Press enter to quit.")



import io
from testpath import assert_isfile, assert_not_path_exists
from zipfile import ZipFile

from nsist import commands

def test_prepare_bin_dir(tmp_path):
    cmds = {
        'acommand': {
            'entry_point': 'somemod:somefunc',
            'extra_preamble': io.StringIO(u'import extra')
        }
    }
    commands.prepare_bin_directory(tmp_path, cmds)

    exe_file = str(tmp_path / 'acommand.exe')

    assert_isfile(exe_file)

    with open(commands.find_exe(console=True), 'rb') as lf:
        b_launcher = lf.read()
        assert b_launcher[:2] == b'MZ'  # Sanity check

    with open(exe_file, 'rb') as ef:
        b_exe = ef.read()
        assert b_exe[:len(b_launcher)] == b_launcher
        assert b_exe[len(b_launcher):].startswith(b"#!<launcher_dir>\\..\\Python\\python.exe\r\n")

    with ZipFile(exe_file) as zf:
        assert zf.testzip() is None
        script_contents = zf.read('__main__.py').decode('utf-8')
    assert 'import extra' in script_contents
    assert 'somefunc()' in script_contents

def test_prepare_bin_dir_noconsole(tmp_path):
    cmds = {
        'acommand': {
            'entry_point': 'somemod:somefunc',
            'console': False
        }
    }
    commands.prepare_bin_directory(tmp_path, cmds)

    exe_file = str(tmp_path / 'acommand.exe')

    assert_isfile(exe_file)

    with open(commands.find_exe(console=False), 'rb') as lf:
        b_launcher = lf.read()
        assert b_launcher[:2] == b'MZ'  # Sanity check

    with open(exe_file, 'rb') as ef:
        b_exe = ef.read()
        assert b_exe[:len(b_launcher)] == b_launcher
        assert b_exe[len(b_launcher):].startswith(b"#!<launcher_dir>\\..\\Python\\pythonw.exe\r\n")

    with ZipFile(exe_file) as zf:
        assert zf.testzip() is None
        script_contents = zf.read('__main__.py').decode('utf-8')
    assert 'somefunc()' in script_contents



import os
import sys

import pytest
from testpath import assert_isfile, assert_isdir

from nsist.copymodules import copy_modules, ExtensionModuleMismatch
from .utils import test_dir, skip_on_windows, only_on_windows

pjoin = os.path.join
running_python = '.'.join(str(x) for x in sys.version_info[:3])

sample_path = [pjoin(test_dir, 'sample_pkgs'),
               pjoin(test_dir, 'sample_zip.egg'),
               pjoin(test_dir, 'sample_zip.egg/rootdir'),
              ]


def test_copy_plain(tmpdir):
    tmpdir = str(tmpdir)
    copy_modules(['plainmod', 'plainpkg'], tmpdir, '3.3.5', sample_path)
    assert_isfile(pjoin(tmpdir, 'plainmod.py'))
    assert_isdir(pjoin(tmpdir, 'plainpkg'))

@skip_on_windows
def test_copy_wrong_platform(tmpdir):
    tmpdir = str(tmpdir)
    with pytest.raises(ExtensionModuleMismatch, match="will not be usable on Windows"):
        copy_modules(['unix_extmod'], tmpdir, '3.3.5', sample_path)

    with pytest.raises(ExtensionModuleMismatch, match="will not be usable on Windows"):
        copy_modules(['unix_extpkg'], tmpdir, '3.3.5', sample_path)

@only_on_windows
def test_copy_windows(tmpdir):
    tmpdir = str(tmpdir)
    copy_modules(['win_extmod', 'win_extpkg'], tmpdir, running_python, sample_path)
    assert_isfile(pjoin(tmpdir, 'win_extmod.pyd'))
    assert_isdir(pjoin(tmpdir, 'win_extpkg'))

@only_on_windows
def test_copy_wrong_pyversion(tmpdir):
    tmpdir = str(tmpdir)
    with pytest.raises(ExtensionModuleMismatch, match="on Python 4"):
        copy_modules(['win_extpkg'], tmpdir, '4.0.0', sample_path)

    with pytest.raises(ExtensionModuleMismatch, match="on Python 4"):
        copy_modules(['win_extmod'], tmpdir, '4.0.0', sample_path)

def test_copy_from_zipfile(tmpdir):
    tmpdir = str(tmpdir)
    copy_modules(['zippedmod2', 'zippedpkg2'],
                 tmpdir, running_python, sample_path)
#        assert_isfile(pjoin(tmpdir, 'zippedmod.py'))
#        assert_isdir(pjoin(tmpdir, 'zippedpkg'))
    assert_isfile(pjoin(tmpdir, 'zippedmod2.py'))
    assert_isdir(pjoin(tmpdir, 'zippedpkg2'))

def test_module_not_found(tmpdir):
    tmpdir = str(tmpdir)
    with pytest.raises(ImportError):
        copy_modules(['nonexistant'], tmpdir, '3.3.5', sample_path)



import sys
from PyQt5 import QtWidgets

from .main import Ui_MainWindow

class Main(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        
        self.ui.add_button.clicked.connect(self.add_item)

    def get_radio_option(self):
        if self.ui.radio_1.isChecked():
            return 'Thing 1'
        elif self.ui.radio_2.isChecked():
            return 'Thing 2'
        elif self.ui.radio_3.isChecked():
            return 'Thing 3'
        elif self.ui.radio_4.isChecked():
            return 'Last thing'
        return 'No thing'
        
    def add_item(self):
        text = self.get_radio_option()
        QtWidgets.QListWidgetItem(text, self.ui.listWidget)        

def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Main()
    window.show()
    sys.exit(app.exec_())



import distlib.scripts
import io
import os.path as osp
import shutil
from zipfile import ZipFile


SCRIPT_TEMPLATE = u"""# -*- coding: utf-8 -*-
import sys, os
import site
installdir = os.path.dirname(os.path.dirname(sys.executable))
pkgdir = os.path.join(installdir, 'pkgs')
sys.path.insert(0, pkgdir)
# Ensure .pth files in pkgdir are handled properly
site.addsitedir(pkgdir)
os.environ['PYTHONPATH'] = pkgdir + os.pathsep + os.environ.get('PYTHONPATH', '')

# Allowing .dll files in Python directory to be found
os.environ['PATH'] += ';' + os.path.dirname(sys.executable)

{extra_preamble}

if __name__ == '__main__':
    from {module} import {func}
    sys.exit({func}())
"""

def find_exe(bitness=32, console=True):
    distlib_dir = osp.dirname(distlib.scripts.__file__)
    name = 't' if console else 'w'
    return osp.join(distlib_dir, '{name}{bitness}.exe'.format(name=name, bitness=bitness))

def prepare_bin_directory(target, commands, bitness=32):
    for name, command in commands.items():
        exe_path = target / (name + '.exe')
        console = command.get('console', True)

        # 1. Get the base launcher exe from distlib
        with open(find_exe(bitness, console=console), 'rb') as f:
            launcher_b = f.read()

        # 2. Shebang: Python executable to run with
        # shebangs relative to launcher location, according to
        # https://bitbucket.org/vinay.sajip/simple_launcher/wiki/Launching%20an%20interpreter%20in%20a%20location%20relative%20to%20the%20launcher%20executable
        if console:
            shebang = b"#!<launcher_dir>\\..\\Python\\python.exe\r\n"
        else:
            shebang = b"#!<launcher_dir>\\..\\Python\\pythonw.exe\r\n"

        # 3. The script to run, inside a zip file
        specified_preamble = command.get('extra_preamble', None)
        if isinstance(specified_preamble, str):
            # Filename
            extra_preamble = io.open(specified_preamble, encoding='utf-8')
        elif specified_preamble is None:
            extra_preamble = io.StringIO()  # Empty
        else:
            # Passed a StringIO or similar object
            extra_preamble = specified_preamble
        module, func = command['entry_point'].split(':')
        script = SCRIPT_TEMPLATE.format(
            module=module, func=func,
            extra_preamble=extra_preamble.read().rstrip(),
        )

        zip_bio = io.BytesIO()
        with ZipFile(zip_bio, 'w') as zf:
            zf.writestr('__main__.py', script.encode('utf-8'))

        # Put the pieces together
        with exe_path.open('wb') as f:
            f.write(launcher_b)
            f.write(shebang)
            f.write(zip_bio.getvalue())

